/*
  Copyright (C) 2001 Jesper Hansen <jesperh@telia.com>.

  Rewritten by: Nikolai Vorontsov <nickviz@mail.be>

  This file is part of the yampp system.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*
	PIN assignements on test board
	
	PA0-PA7		data bus 0..7 + Address 0..7
	PC0-PC7		address 8..15	
	
	PB0		T0		DIOW		ATA out
	PB1		T1		DIOR		ATA out
	PB2		AIN0	DREQ		VS1001 in		
	PB3		AIN1	BSYNC		VS1001 out
	PB4		SS		MP3			VS1001 out
	PB5		MOSI	SO			VS1001 out
	PB6		MISO 	SI			VS1001 in
	PB7		SCK		SCK			VS1001 out

	PD0		RxD		RS-232		UART in
	PD1		TxD		RS-232		UART out
	PD2		INT0	IR_INPUT	IR in
	PD3		INT1	KEY_INPUT	KEY in
	PD4		T0		
	PD5		T1		LCD_ENABLE	LCD out
	PD6		WR		WR			RAM out
	PD7		RD		RD			RAM out
*/


#include <stdlib.h>
#include <io.h>
#include <progmem.h>
#include <eeprom.h>
#include <sig-avr.h>
#include <interrupt.h>
#include <wdt.h>

#include "Constants.h"
#include "types.h"
#include "mem.h"
#include "delay.h"
#include "ata_if.h"
#include "fat.h"
#include "vs1001.h"
#include "rec80.h"
#include "uart.h"
#include "lcd.h"


// Amount of small steps in LCD play position bar
#define LCD_STEPS 50

// Amount of buffers writed before relinquish control
#define MAX_CONS_WRITES 128

typedef enum
{
	EV_IDLE,
	EV_PLAY,
	EV_STOP,
	EV_PAUSE,
	EV_NEXT,
	EV_PREV,
	EV_NEXT10,
	EV_PREV10,
	EV_VOLUP,
	EV_VOLDN,
	EV_NEUTRAL,
	EV_MUTE,
	EV_RANDOM,
	EV_NEXTL,
	EV_PREVL,
	EV_LOUDNESS,
	EV_LASTEVENT	
} event_e;	


// Auxilliary functions declarations
void sectordump(u32 sector);
u08* unsigned2str(u16 track, u08 *str);
event_e get_char_event(void);
event_e ir_event(void);
event_e get_key_event(void);
void send_sinewave_beeps(void);
void userchars(void);
void setvolume(u08 v);
void dispbar(void);
bool RetrieveMP3Data(void);


// Global variables
u08 scroll_length	= 0;
u08 scroll_pos		= 0;
char *scroll_line	= 0;


bool bPlaying		= false;		// Current state - play/stop
u08 *pPlayData		= 0;			// Pointer to current position in currently
									//  used buffer
u08 *pPlayBufStart	= 0;			// Pointer to currenly used buffer
u32 dwPlayed		= 0;			// Amount of already played bytes

// timer 1 routines
//
//	At 8MHz clock :			100 mS divisor		8bit overflow
//	tClk 		= 125 nS 	   200000				32 uS (31250 Hz)
//	tClk/8 		=   1 uS		50000 			   128 uS (7812.5 Hz)
//	tClk/64		=   8 uS	 	12500			   512 us (1953.125 Hz)
//	tClk/256	=  32 uS		 3125	 		  2048 uS
//	tClk/1024	= 128 uS	 	  781.25		  8192 uS
//
// prescaler set to 32 uS
// 3125 ticks = 100 mS (for 8MHz!)

#define TI1_H	(((u16)-(F_CPU / 2560)) >> 8)
#define TI1_L	(((u16)-(F_CPU / 2560)) & 0xff )

volatile u08 nDispUpd = 0;			// Counter for display update loop
volatile u08 nScrollUpd = 0;		// Counter for name scroll loop
volatile u08 nKeyTime = 0;			// Counter for keypress time check - Added by MIS

SIGNAL(SIG_OVERFLOW1)				// Timer1 overflow every 100 mS
{
	nDispUpd++;
	nScrollUpd++;
	nKeyTime++;
	outp(TI1_H, TCNT1H);			// Reload timer
	outp(TI1_L, TCNT1L);
}


// Random stuff, maximum 65535 songs on the disk!
#define MY_RAND_MAX  	65535
static u32 seed = 1;

u16 do_rand(u16 max)				// returned random value from 0 to max - 1
{
	return ((seed = seed * 1103515245L + 12345) % max);
}


u08 *randstr  = "Rand*";
u08 *loudstr  = "Loud*";
u08 trackstr[6], *pTrack, timestr[6], tot_sw = 0;

//----------------------------------------------------------------------------
// Main part
//----------------------------------------------------------------------------
int main(void)
{
	bool bAutoPlay = true, bRandom = true;	// Modificators
	bool bLoudness = true;					// Modificators
	u08 nVolume = 0;						// Current volume level
	u16 wSongNum = 0;						// Current playing song
	u16 wPlayTime;							// Current play time
	u16 wMaxSongs = 1;						// Maximum available songs
	u08 nWrites, nBursts;					// Buffering variables
	event_e event = EV_IDLE;				// Current event
	u08 nTemp, *src, *dst, *pTemp;			// Temporary variables
	u16 i;
	udiv_t divt;

//----------------------------------------------------------------------------
// B - Port
//----------------------------------------------------------------------------
	sbi(PORTB, 0);		// DIOW- hi
	sbi(DDRB,  0);		// Pin PB0 is output, DIOW- 
	sbi(PORTB, 1);		// DIOR- hi
	sbi(DDRB,  1);		// Pin PB1 is output, DIOR- 
	cbi(DDRB,  2);		// Pin PB2 is input, DEMAND
	sbi(PORTB, 2);		// Activate pullup
	// PB3 and PB4 is used for BSYNC and MP3 signals to VS1001
	// PB5 (MOSI) is used for SPI
	// PB6 (MISO) is used for SPI
	// PB7 (SCK) is used for SPI

//----------------------------------------------------------------------------
// D - Port
//----------------------------------------------------------------------------
	// PD0 and PD1 are TX and RX signals
	// PD2 is used for the IR receiver
	cbi(DDRD, 3);		// Pin PD3 is keyboard input
	// PD4 is available, but activate pullup to aviod noice problem
	sbi(PORTD, 4);	
	// PD5 is LCD_ENABLE used by LCD module
	// PD6 and PD7 are RD and WR signals
	// but we need to set their default states used
	// when the external memory interface is disabled
    // they should then both be high to avoid disturbing
	// RAM contents
	sbi(DDRD,  6);		// As outputs
	sbi(DDRD,  7);
	sbi(PORTD, 6);		// And both hi
	sbi(PORTD, 7);

//----------------------------------------------------------------------------
// Timer 1
//----------------------------------------------------------------------------
	// setup timer1 for a 100mS periodic interrupt
	outp(0, TCCR1A);
	outp(4, TCCR1B);				// Prescaler /256  tPeriod = 32 uS
	outp(TI1_H, TCNT1H);			// Load counter value hi
	outp(TI1_L, TCNT1L);			// Load counter value lo

//----------------------------------------------------------------------------
// Initialization
//----------------------------------------------------------------------------
    lcd_init(0, LCD_FUNCTION_DEFAULT);
	userchars();		// Construct some user definable characters
	lcd_clrscr();
	lcd_puts(" Hi, MP3 world!\n yampp-3 player");

	UART_Init();		// Init RS-232 link
	vs1001_init_io();	// Init IO interface for VS1001
	vs1001_reset(false);	// Init VS1001
	init_rec80();		// Set up IR receiver interface

	// Say hello to UART
 	PRINT("\n\nyampp-3 "__DATE__" "__TIME__"\nATA checking...");
						
	ATA_Init();			// Init and wait for drive
	ATA_SW_Reset();		// Reset drive
	PRINT("done.\n");

#ifdef ENABLE_WATCHDOG	// Enable watchdog (enabled by linker option)
	wdt_enable(0x07);
#endif

	seed = 1 + inp(TCNT1L);			// Init randomizer from timer 
									// (ATA init gives us random delay)

	if (!init_fat(DRIVE0))			// Read disk structure
	{
		lcd_clrscr();
		lcd_puts("Wrong disk!");
		while (true) WDR;			// Lock player
	}

	PRINT("Scanning...");
//	wMaxSongs = dirlist(DIRLIST_VERBOSE);
	wMaxSongs = dirlist(DIRLIST_SCAN);

	PRINT(unsigned2str(wMaxSongs , trackstr));	// jman number of songs in decimal
	PRINT(" entries\n");

	nVolume = eeprom_rb(EEPROM_VOLUME);			// Load and set initial volume
	if (nVolume > MIN_VOLUME)
		nVolume = MIN_VOLUME;
	setvolume(nVolume);

	send_sinewave_beeps();			// Send three sinewave beeps to indicate
									// we're starting up okay
	// load bAutoPlay and position settings
	bAutoPlay = eeprom_rb(EEPROM_AUTOPLAY);		// Load autoplay and
	bRandom	  = eeprom_rb(EEPROM_RANDOM);		//  position settings
	wSongNum  = eeprom_rw(EEPROM_SONGPOS);	
	bLoudness = eeprom_rb(EEPROM_LOUDNESS);	

	// set initial status in the strings
	*(randstr+4) = bRandom ? '+' : '-';
	*(loudstr+4) = bLoudness ? '+' : '-';

	if (bAutoPlay)
		event = EV_PLAY;

//----------------------------------------------------------------------------
// start things rolling
//----------------------------------------------------------------------------
	sbi(TIMSK, TOIE1);	 			// Enable timer1 interrupt
	while (true)					// Main loop
	{
		WDR;						// Give watchdog full time
		if (UART_HasChar())			// Check all input controls
			event = get_char_event();
		if (rec80_active())
			event = ir_event();
		if (event == EV_IDLE)
			event = get_key_event();
			
		if (bPlaying && event == EV_PLAY)	// second play press does pause
			event = EV_PAUSE;

		switch (event)				// According input control
		{
			case EV_IDLE:
				break;
#ifdef ENABLE_NAV
			case EV_PREVL:			// Switch to previous start letter 
				nTemp = nLastPos - 1;
				while (nTemp < nLastPos)
					if (LetterTable[nTemp].letter < *LONG_NAME)
						break;
					else
						nTemp--;

				if (nTemp >= nLastPos)
					nTemp = nLastPos - 1;
				wSongNum = LetterTable[nTemp].song_num;
				goto PLAY;
			case EV_NEXTL:			// Switch to next start letter
				for (nTemp = 1; nTemp < nLastPos; nTemp++)
					if (LetterTable[nTemp].letter > *LONG_NAME)
						break;
				if (nTemp >= nLastPos)
					nTemp = 0;
				wSongNum = LetterTable[nTemp].song_num;
				goto PLAY;
#endif

			case EV_PREV10:	wSongNum -= 10;	goto PLAY;
			case EV_NEXT10:	wSongNum += 10;	goto PLAY;
			case EV_PREV:
				if (wSongNum == 0)
					wSongNum = wMaxSongs - 1;
				else
					wSongNum--;
				goto PLAY;
			case EV_NEXT: wSongNum++;
				// fall through
			case EV_PLAY:
PLAY:			lcd_clrscr();
				lcd_puts("Loading...");
				vs1001_reset(bLoudness);		// Mandatory!

				if (wSongNum >= wMaxSongs)		// Protect for another disk
					wSongNum = 0;
				eeprom_ww(EEPROM_SONGPOS, wSongNum);

				if (!get_dir_entry(wSongNum))	// Setup buffer for play
				{
					wSongNum = 0;				// Something wrong
					continue;
				}
				PRINT("\nPlaying #"); 
				PRINT(pTrack = unsigned2str(wSongNum + 1, trackstr)); 
				UART_SendByte(' ');	PRINT(LONG_NAME);

				scroll_line = src = LONG_NAME;
				scroll_pos	= 0;
				scroll_length = 4;				// To add ' >> ' sign

				while (*src++)
					scroll_length++;
				src--;
				// *((u32*)p)++ = 0x207E7F20L;
				*((u32*)src)++ = 0x203E3E20L;
				*src = 0;

				dwPlayed = 0;
				nScrollUpd = 0;
				nDispUpd = 5;
				tot_sw = 10;				
				ReadBuffer(BUF1);				// Preload buffers
				if (CheckAndAdvanceCluster())	// Long enough file
					ReadBuffer(BUF2);
				pPlayData = pPlayBufStart = BUFFER1P;
				bPlaying = true;

				// Special MP3 header patch to strip all invalid headers 
				// not needed for VS1001K chip
/*				for (i = 0; i < BUFFER_SIZE - 1; i++)
					if ((*(u16*)&BUFFER1P[i] & 0xfeff) != 0xfaff)
						BUFFER1P[i] = 0;
					else
						break;
				// End of special header patch
*/				
				break;
			case EV_STOP:
				bPlaying = false;
				lcd_clrscr();
				lcd_puts("yampp-3");
				break;
			case EV_VOLDN:
				if (nVolume <= MIN_VOLUME-VOL_STEP) 
				{
					nVolume += VOL_STEP;
					setvolume(nVolume);
				}
				else
					setvolume(0xfe);
				break;
			case EV_VOLUP:
				if (nVolume >= VOL_STEP) 
					nVolume -= VOL_STEP;
				setvolume(nVolume);
				break;
			case EV_NEUTRAL:
				nVolume = 12;		// -6 dB
				setvolume(nVolume);
				break;
//			case EV_MUTE:
//				mute = 1-mute;
//				break;
			case EV_LOUDNESS:
				bLoudness = !bLoudness;
				*(loudstr+4) = bLoudness ? '+' : '-';
				vs1001_write(VS1001_MODE, 0x0080 * (bLoudness == true)); // write bit SM_BASS to VS1001K
				eeprom_wb(EEPROM_LOUDNESS, bLoudness);
				nDispUpd = 5;
				tot_sw = 14;				
				break;
				
			case EV_RANDOM:
				bRandom = !bRandom;
				*(randstr+4) = bRandom ? '+' : '-';
				eeprom_wb(EEPROM_RANDOM, bRandom);
				nDispUpd = 5;
				tot_sw = 12;				
				break;

//jesper 011207	start			
			case EV_PAUSE:
				event = EV_IDLE;
				lcd_gotoxy(10,1);
				lcd_puts("Paused");
				while (rec80_active()) ;	// wait for key release
				do
				{
					WDR;						// Give watchdog full time
					if (UART_HasChar())			// Check all input controls
						event = get_char_event();
					if (rec80_active())
						event = ir_event();
					if (event == EV_IDLE)
						event = get_key_event();
				} while (event == EV_IDLE);
				break;
//jesper 011207 end	
				
			default:
				break;
		} // switch (event)

		event = EV_IDLE;			// We always go back to IDLE when
									//  exiting switch
									
									
									
									
									
		if (bPlaying)
		{
			WDR;					// Enable time to push data

			nWrites = 0;			// This count prevents indefinite
			do						//  feed of data to VS1001 with no
			{						//  way to get control
									// Amount of rest of data bursts
				nBursts = (u08)(((pTemp = pPlayData) - pPlayBufStart) / 32);
				while (nBursts < BUFFER_SIZE / 32 &&// While we have data and 
					   bit_is_set(PINB, PB2))		//  DREQ is high
				{
					vs1001_send_32(pPlayData);		// Feed data to VS1001
					pPlayData += 32;				// Advance pointer
					nBursts++;
				}
				dwPlayed += pPlayData - pTemp;		// Count already pumped bytes
			} while(++nWrites < MAX_CONS_WRITES  &&	// We write not too many times
					nBursts == BUFFER_SIZE / 32  &&	//  and we didn't stop by DREQ
					RetrieveMP3Data());				//  and we read more data

			if (bBuf1Empty && bBuf2Empty)			// No more data to play
			{										// Try to autoplay next song
				bPlaying = false;
#ifdef ENABLE_CRC
				UART_Printfu16(crc);
#endif
				if (bRandom)
				{
					wSongNum = do_rand(wMaxSongs);
					event = EV_PLAY;
				}
				else
					event = EV_NEXT;
				continue;
			}
			
			if (nScrollUpd >= 5)	// Scroll song name every 500ms
			{
				lcd_gotoxy(0, 0);
				for (nTemp = 0; nTemp < 16; nTemp++)
					lcd_putchar(scroll_line[(scroll_pos + nTemp) % scroll_length]);
				scroll_pos++;
				nScrollUpd = 0;
			}

			if (nDispUpd >= 5)		// Update other display info every 500ms
			{
				dispbar();

				lcd_gotoxy(10,1);

				dst = timestr;
				
				if (tot_sw < 10)	// 0-9
				{
					src = timestr;
					wPlayTime = vs1001_read(VS1001_PLAYTIME);
					
					divt = udiv(wPlayTime, 3600);
					divt = udiv(divt.rem, 600);
					*src++ = '0' + (u08)divt.quot;
					divt = udiv(divt.rem, 60);
					*src++ = '0' + (u08)divt.quot;
					*src++ = ':';
					divt = udiv(divt.rem, 10);
					*src++ = '0' + (u08)divt.quot;
					*src++ = '0' + (u08)divt.rem;			
					*src = 0;
				}
				else if (tot_sw < 12)	// 10-11
				{
					*dst++ = '#';
					src = pTrack;
					while ((*dst = *src++)) dst++;
					while (dst < &timestr[5])
						*dst++ = ' ';
					dst = timestr;
				}
				else if (tot_sw < 14)	// 12-13
				{
					dst = randstr;
					//while (*dst++ = *src++);
				}
				else if (tot_sw < 16)	// 14-15
				{
					dst = loudstr;
					//while (*dst++ = *src++);
				}
	
				tot_sw += 2;
				
				if (tot_sw >= 16)
					tot_sw = 0;

//				lcd_puts(timestr);
				lcd_puts(dst);
				
				lcd_putchar(0);		// Volume indicator
				nDispUpd = 0;
				
			} // if update time display
		} // if (bPlaying)
	} // while (true)
}




// Auxillary functions definitions

// converts 16 bit unsigned integer to string
// str[] should be 6 bytes at least. Function can return value != str
// and you have to use it as a output value start
u08* unsigned2str(u16 track, u08 *str)
{
	register u08 i;
	udiv_t divt;

	for (i = 0; i < 5; i++) *str++ = ' ';
	*str = '\0';
	divt.quot = track;
	do {
		divt = udiv(divt.quot, 10u);
		*(--str) = '0' + (u08)divt.rem;
	} while (divt.quot);
	return str;
}


#ifdef ENABLE_SERIAL

typedef struct pairUART				// auxilliary structure for making
{									//  of pairs table for select operator
	u08 code, action;
} pairUART;

pairUART TableUART[] = { 
	{ 0, EV_IDLE }, 
	{ 'g', EV_PLAY },	 { 'G', EV_STOP }, 
	{ 'p', EV_PREV },	 { 'n', EV_NEXT }, 
	{ 'P', EV_PREV10 },  { 'N', EV_NEXT10 },
	{ 'V', EV_VOLUP },	 { 'v', EV_VOLDN },
	{ 'u', EV_NEUTRAL }, { 'm', EV_MUTE },
#ifdef ENABLE_NAV
	{ 'l', EV_NEXTL },	 { 'L', EV_PREVL },
#endif
	{ 'r', EV_RANDOM },	 { ' ', EV_PAUSE }
};


event_e get_char_event(void)
{
#ifdef HEXDUMP
	static u32 dwLbaSector = 0;
	static u32 dwTempAcc = 0;
#endif

	register u08 i;
	register u08 code = UART_ReceiveByte();

	for (i = 0; i < sizeof(TableUART) / sizeof(TableUART[0]); i++)
		if (TableUART[i].code == code)
			return TableUART[i].action;
#ifdef HEXDUMP
	if (code >= '0' && code <= '9')
		dwTempAcc = dwTempAcc * 10 + code - '0';
	if (code == 0x0d)		// Enter
	{
		sectordump(dwLbaSector = dwTempAcc);
		dwTempAcc = 0;
	}
	if (code == '+')
		sectordump(++dwLbaSector);
	if (code == '-')
		sectordump(--dwLbaSector);
#endif	// HEXDUMP

    return EV_IDLE;
}

#endif // ENABLE_SERIAL


// get a remote control event
typedef struct pairIR
{
	u16 code;
	u08 action;
} pairIR;

pairIR TableIR[] = { 
	{ 0, EV_IDLE },
	{ IR_PLAY, EV_PLAY },	  { IR_STOP, EV_STOP }, 
	{ IR_PREV, EV_PREV },	  { IR_NEXT, EV_NEXT }, 
	{ IR_PREV10, EV_PREV10 }, { IR_NEXT10, EV_NEXT10 },
	{ IR_VOLUP, EV_VOLUP },   { IR_VOLDN, EV_VOLDN },
#ifdef ENABLE_NAV
	{ IR_NEXTL, EV_NEXTL },	  { IR_PREVL, EV_PREVL },
#endif
	{ IR_RANDOM, EV_RANDOM }, { IR_PAUSE, EV_PAUSE }
};

event_e ir_event(void)
{
	register u08 i;
	u16 code = get_rec80();
	for (i = 0; i < sizeof(TableIR) / sizeof(TableIR[0]); i++)
		if (TableIR[i].code == code)
			return TableIR[i].action;

	return EV_IDLE;
}



// get a keyboard event - edhanced by MIS

u08 TableKbdShort[] = { EV_IDLE, EV_LOUDNESS, EV_PLAY, EV_PREV,   EV_NEXT,   EV_PREVL, EV_NEXTL, EV_VOLUP, EV_VOLDN };
u08 TableKbdLong[]  = { EV_IDLE, EV_RANDOM,   EV_STOP, EV_PREV10, EV_NEXT10, EV_PREVL, EV_NEXTL, EV_VOLUP, EV_VOLDN };

event_e get_key_event(void)
{
	static u08 keydown = 0;
	register u08 i;
	event_e event = EV_IDLE;

	cbi(MCUCR, SRE);		// disable ExtRAM

	outp(0xff, PORTA-1);	// set port as output
	outp(0, PORTA);			// set all pins to 0
	asm volatile("nop");	// allow time for port to activate

	if (keydown)			// if a key was pressed earlier
	{
		// see if it has been released
		if (bit_is_set(PIND, PD3))				// check bit
		{
			// no key is down, return last key seen
			if (nKeyTime < 8)					// max 0.8 sec for shortpress
				event = TableKbdShort[keydown];	// short keypress event
			keydown = 0;
		}
		else
			if ((nKeyTime & 0x7F) > 8)			// long keypress ? 
			{
				event = TableKbdLong[keydown];	// long keypress event
				nKeyTime = 0x83;				// 0.6 sec repeat time
			}

	}
	else
	{	
		if (bit_is_clear(PIND, PD3))	// check if a key is down
		{
			// a key is active, check which one
			for (i = 0; i < 8; i++)
			{
				outp(~(1 << i), PORTA);			// write bit mask
				asm volatile("nop");			// wait a while
				if (bit_is_clear(PIND, PD3)) 	// this bit ?
				{
					keydown = i + 1;	// store the key
					nKeyTime = 0;		// clear keypress timer
					break;				// and exit
				}
			}
		}
	}

	sbi(MCUCR, SRE);					// enable ExtRAM
	return event;
}



void send_sinewave_beeps(void)
{
	static u08 buf[] = { 0x53, 0xEF, 0x6E , 0x30, 0x45, 0x78, 0x69, 0x74 };
	register u08 i, j = 3;

	while (j-- != 0)
	{
		for (i = 0; i < 4; i++)
			vs1001_send_data(buf[i]);
		vs1001_nulls(4);
		delay10(10000);		// delay 100ms

		for (/*i = 4*/; i < 8; i++)
			vs1001_send_data(buf[i]);
		vs1001_nulls(4);
		delay10(10000);
	}	
}


void userchars(void)
{
#ifdef ENABLE_LCD
	static u08 data[] = { 0x10, 0x18, 0x1C, 0x1E, 0x1F };
	register u08 i = 0;
	
	lcd_command(0x48);		// start at 2'nd definable character
	while (i < 40)
		lcd_data(data[i++ >> 3]);	// eight times send each value
	// 2 more user definable characters possible
#endif
}


void setvolume(u08 v)
{
	register u08 i;

	vs1001_setvolume(v, v);
	eeprom_wb(EEPROM_VOLUME, v);

	lcd_command(0x40);

	for (i = 0; i < 8; i++)
	{
		if (v <= i * (VOL_STEP*(NUM_VOL_STEPS/8)))
			lcd_data((1 << ((10 - i) / 2)) - 1);	// Volume idicator mod
		else
			lcd_data(0x00);
	}
}


void dispbar(void)					// 10 bars of 5 steps each = 50 steps
{
	register u08 i, a, nSteps = (u08)((dwPlayed * LCD_STEPS) / dwFileSize);

	a = nSteps / 5;						// # of full bars

	lcd_gotoxy(0, 1);				

	for (i = 0; i < a; i++)
		lcd_putchar(5);

	if (a <= (LCD_STEPS / 5) - 1)	// jman to fix display overwrite
		lcd_putchar(1 + nSteps % 5);	
}


void eeprom_ww(u08 address, u16 value)
{
	union u16convert Data;
	Data.value = value;

	while (!eeprom_is_ready());
	eeprom_wb(address, Data.bytes.low);
	while (!eeprom_is_ready());
	eeprom_wb(address + 1, Data.bytes.high);
	while (!eeprom_is_ready());
}


bool RetrieveMP3Data(void)
{
// Buf1 Buf2 Action
//	E    E   Read full, return true
//	F    E   if (prep_read)
//	E    F		return true
//			 else
//			    read_info
//	F    F   prep_read
	u08 nRes;

	if (pPlayBufStart == BUFFER1P)	// Check that we exhausted buffer
	{
		if (pPlayData == BUFFER1P + BUFFER_SIZE)
			bBuf1Empty = true;
	}
	else							// pPlayBufStart == BUFFER2P
	{
		if (pPlayData == BUFFER2P + BUFFER_SIZE)
			bBuf2Empty = true;
	}

	if (bBuf1Empty  &&  bBuf2Empty)	// First case - both buffers empty
	{								//  unconditional read
		if (CheckAndAdvanceCluster())
		{
			ReadBuffer(BUF1);
			pPlayData = pPlayBufStart = BUFFER1P;
			return true;			// Need imeediate pump in
		}
		else
			return false;			// No more data for this file
	}

	if (!bBuf1Empty  &&  !bBuf2Empty)// Both buffers filled, refresh FAT cache 
		return CheckAndAdvanceCluster() == FAT_READ;

	if (bBuf1Empty)
	{
		if (pPlayBufStart == BUFFER1P)		// Need to switch buffer
			pPlayData = pPlayBufStart = BUFFER2P;
		nRes = CheckAndAdvanceCluster();
		if (nRes == NO_DATA)
			return false;
		if (nRes == DATA_EXIST)
			ReadBuffer(BUF1);
		return true;				// Was disk operation, need pump in
	}

	if (bBuf2Empty)
	{
		if (pPlayBufStart == BUFFER2P)		// Need to switch buffer
			pPlayData = pPlayBufStart = BUFFER1P;
		nRes = CheckAndAdvanceCluster();
		if (nRes == NO_DATA)
			return false;
		if (nRes == DATA_EXIST)
			ReadBuffer(BUF2);
		return true;				// Was disk operation, need pump in
	}
}


#ifdef HEXDUMP

void HexDump(u08 *buffer, u16 len)
{
    u16 i, j, k;

    EOL();
    for (i = 0, k = 0; i < len / 16; i++, k += 16)
    {
        UART_Printfu16(k);
        UART_SendByte(' ');
        for (j = 0; j < 16; j++)
        {
            UART_Printfu08(buffer[i * 16 + j]);	
            UART_SendByte(' ');
        }
        PRINT("    ");
        for(j =0 ; j < 16; j++)
        {
            if (buffer[i * 16 + j] >= 0x20)
                UART_SendByte(buffer[i * 16 + j]);
            else
                UART_SendByte('.');
        }
        EOL();
    }
}

void sectordump(u32 sector)
{
	if (ATA_ReadLBA0(sector, 1, TMP_SECTOR) == 0)
	{
		UART_Printfu32(sector);
		HexDump(TMP_SECTOR, BPS);
	}
	else
		PRINT("Dump failed\n");
}

#endif

